/* macio.c: Probe for i2o audio and airport on Apple machines.
 * 
 * Copyright (C) 2002 Dan Burcaw <dburcaw@terrasoftsolutions.com>
 *           (C) 2002 Terra Soft Solutions, Inc.
 *           (C) 2003 Red Hat, Inc.
 *
 * Based on sbus.c
 * Copyright (C) 1998, 1999 Jakub Jelinek (jj@ultra.linux.cz)
 *           (C) 1999, 2000 Red Hat, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "macio.h"

static void macioFreeDevice( struct macioDevice *dev ) {
    freeDevice( (struct device *)dev);
}

static void macioWriteDevice( FILE *file, struct macioDevice *dev) {
    writeDevice(file, (struct device *) dev);
}

static int macioCompareDevice(struct macioDevice *dev1, struct macioDevice *dev2)
{
    return compareDevice( (struct device *)dev1, (struct device *)dev2);
}

struct macioDevice *macioNewDevice( struct macioDevice *old ) {
    struct macioDevice *ret;

    ret = malloc( sizeof(struct macioDevice) );
    memset(ret, '\0', sizeof (struct macioDevice));
    ret=(struct macioDevice *)newDevice((struct device *)old,(struct device *)ret);
    ret->bus = BUS_MACIO;
    ret->newDevice = macioNewDevice;
    ret->freeDevice = macioFreeDevice;
    ret->writeDevice = macioWriteDevice;
    ret->compareDevice = macioCompareDevice;
    return ret;
}

#ifdef __powerpc__
#include "minifind.h"
static int check_of_compatible(char *path, char **matches)
{
	int i, fd;
	size_t size;
	char buf[256];
	char *str = buf;

	buf[255] = 0;
	snprintf(buf, 255, "%s/compatible", path);

	fd = open(buf, O_RDONLY);
	if (fd < 0)
		return 0;

	size = read(fd, buf, 255);
	if (size < 0) {
		close(fd);
		return 0;
	}
	close(fd);

	while (str < &buf[size]) {
		for (i=0; matches[i]; i++) 
			if (!strcmp(matches[i], str))
				return i+1;
		str += strlen(str) + 1;
	}
	return 0;
}

static void free_find_results(struct findNode *list)
{
	while (list->result) {
		struct pathNode *tmp = list->result;
		list->result = tmp->next;
		free(tmp);
	}	
}

static void find_of_netdev(struct macioDevice *netdev, char *path)
{
	char buf[256];
	struct dirent **namelist;
	struct stat statbuf;
	int n, fd;
	size_t sz;

	buf[255] = 0;

	path += strlen("/proc/device-tree");

	n = scandir("/sys/class/net", &namelist, 0, 0);
	if (n >= 0) {
		while (n--) {
			snprintf(buf, 255, "/sys/class/net/%s/device/devspec", namelist[n]->d_name);
			fd = open(buf, O_RDONLY);
			if (fd >= 0) {
				sz = read(fd, buf, 255);
				if (sz >= 0) {
					buf[sz] = 0;
					if (!strcmp(buf, path)) {
						netdev->device = strdup(namelist[n]->d_name);
						__getNetworkAddr((struct device *)netdev, netdev->device);
						while (n--)
							free(namelist[n]);
						return;
					}
				}
			}
			free(namelist[n]);
		}
	}

	netdev->device = strdup("eth");
}


/* given a class to probe, returns an array macio devices found which match */
/* all entries are malloc'd, so caller must free when done. Use          */
/* dev->freeDevice(dev) to free individual macioDevice's, since they 	 */
/* have internal elements which are malloc'd                             */

struct device *macioProbe( enum deviceClass probeClass, int probeFlags,
		    struct device *devlist) {

	struct findNode list;
	struct pathNode *n;

	memset(&list, 0, sizeof(list));

	if (probeClass & CLASS_USB) {
		struct macioDevice *ohci;

		minifind("/proc/device-tree", "usb", &list);

		// Supported
		for (n = list.result; n; n = n->next) {
			char *matches[] = {"ohci-le", "ohci-littleendian", "ohci-be",
					   "ohci-bigendian", "mpc5200-ohci",
					   "mpc5200-usb-ohci", NULL};

			if (check_of_compatible(n->path, matches)) {
				ohci = macioNewDevice(NULL);
				ohci->type = CLASS_USB;
				ohci->desc = strdup("OpenFirmware OHCI");
				ohci->driver = strdup("ohci-hcd");
				ohci->next = devlist;
				devlist = (struct device *)ohci;
				break;
			}
		}
		free_find_results(&list);
	}

	// check for airport
	if (probeClass & CLASS_NETWORK) {
		struct macioDevice *netdev;
		struct pathNode *n;

		minifind("/proc/device-tree", "radio", &list);

		// Supported
		if (list.result) {
			netdev = macioNewDevice(NULL);
			netdev->type = CLASS_NETWORK;
			find_of_netdev(netdev, n->path);
			netdev->desc = strdup("Apple Computer Inc.|Airport");
			netdev->driver = strdup("airport");
			netdev->next = devlist;
			devlist = (struct device *)netdev;
		}
		free_find_results(&list);

		minifind("/proc/device-tree", "ethernet", &list);

		// Supported
		for (n = list.result; n; n = n->next) {
			int ret;
			char *matches[] = {"bmac", "bmac+", "mpc5200-fec", NULL};

			ret = check_of_compatible(n->path, matches);
			if (!ret)
				continue;

			netdev = macioNewDevice(NULL);
			netdev->type = CLASS_NETWORK;
			find_of_netdev(netdev, n->path);
			switch(ret) {
			case 1:
			case 2:
				netdev->desc = strdup("Apple Computer Inc.|BMAC/BMAC+");
				netdev->driver = strdup("bmac");
				break;
			case 3:
				netdev->desc = strdup("Freescale|MPC5200 FEC");
				netdev->driver = strdup("fec_mpc52xx");
				break;
			}
			netdev->next = devlist;
			devlist = (struct device *)netdev;
			break;
		}
		free_find_results(&list);

		/* Yes, this is a platform device */
		minifind("/sys/devices/platform", "mv643xx_eth", &list);

		if (list.result) {
			netdev = macioNewDevice(NULL);
			netdev->type = CLASS_NETWORK;
			__getSysfsDevice((struct device *)netdev, n->path, "net:",0);
			if (netdev->device)
				__getNetworkAddr((struct device *)netdev, netdev->device);
			else
				netdev->device = strdup("eth");
			netdev->desc = strdup("Marvell|MV643xx Ethernet");
			netdev->driver = strdup("mv643xx_eth");
			netdev->next = devlist;
			devlist = (struct device *)netdev;
		}
		free_find_results(&list);

	}

	if (probeClass & CLASS_AUDIO) {
		struct macioDevice *dmasound;
		char buf[256];

		minifind("/proc/device-tree", "sound", &list);

		// Supported
		for (n = list.result; n; n = n->next) {

			if (!strstr(n->path, "mac-io"))
				continue;

			dmasound = macioNewDevice(NULL);
			dmasound->type = CLASS_AUDIO;
			dmasound->next = devlist;

			snprintf(buf, 255, "%s/layout-id", n->path);
			if (!access(buf, O_RDONLY)) {
				dmasound->desc = strdup("Apple Computer Inc.|Apple Onboard Audio");
				dmasound->driver = strdup("snd-aoa-i2sbus");
			} else {
				dmasound->desc = strdup("Apple Computer Inc.|PowerMac Sound");
				dmasound->driver = strdup("snd-powermac");
			}
			devlist = (struct device *)dmasound;
			break;
		}
		free_find_results(&list);
	}

	if (probeClass & CLASS_IDE) {
		struct macioDevice *ide;

		minifind("/proc/device-tree", "ata", &list);

		// Supported
		for (n = list.result; n; n = n->next) {
			char *matches[] = {"mpc5200-ata", NULL};
			if (check_of_compatible(n->path, matches)) {
				ide = macioNewDevice(NULL);
				ide->type = CLASS_IDE;
				ide->next = devlist;
				ide->desc = strdup("Freescale|MPC52xx ATA");
				ide->driver = strdup("pata_mpc52xx");
				devlist = (struct device *)ide;
				break;
			}
		}
		free_find_results(&list);

		minifind("/proc/device-tree", "ide", &list);

		// Supported
		for (n = list.result; n; n = n->next) {
			char *matches[] = {"electra-ide", NULL};
			if (check_of_compatible(n->path, matches)) {
				ide = macioNewDevice(NULL);
				ide->type = CLASS_IDE;
				ide->next = devlist;
				ide->desc = strdup("PA Semi|Electra IDE");
				ide->driver = strdup("electra-ide");
				devlist = (struct device *)ide;
				break;
			}
		}
		free_find_results(&list);
	}

	if (probeClass & CLASS_OTHER) {
		struct macioDevice *i2c;
	    
		char *matches[] = {"PowerMac7,2", "PowerMac7,3", "RackMac3,1", NULL};
		if (check_of_compatible("/proc/device-tree", matches)) {
			i2c = macioNewDevice(NULL);
			i2c->type = CLASS_OTHER;
			i2c->desc = strdup("Apple Computer Inc.|G5 Thermostat");
			i2c->driver = strdup("therm_pm72");
			i2c->next = devlist;
			devlist = (struct device *)i2c;
		}

		minifind("/proc/device-tree", "i2c", &list);
	    
		// Supported
		for (n = list.result; n; n = n->next) {
			int ret;
			char *matches[] = {"keywest-i2c", "hydra-i2c", NULL};

			ret = check_of_compatible(n->path, matches);
			if (!ret)
				continue;

			i2c = macioNewDevice(NULL);
			i2c->type = CLASS_OTHER;
				
			if (ret == 1) {
				i2c->desc = strdup("Apple Computer Inc.|KeyWest I2C");
				i2c->driver = strdup("i2c-keywest");
			} else {
				i2c->desc = strdup("Apple Computer Inc.|Hydra I2C");
				i2c->driver = strdup("i2c-hydra");
			}
			i2c->next = devlist;
			devlist = (struct device *)i2c;
			break;
		}
		free_find_results(&list);

		minifind("/proc/device-tree", "fan", &list);
		// Supported
		for (n = list.result; n; n = n->next) {
			char *matches[] = {"adt7460", "adt7467", NULL};
			if (check_of_compatible(n->path, matches)) {
				i2c = macioNewDevice(NULL);
				i2c->type = CLASS_OTHER;
				i2c->desc = strdup("Apple Computer Inc.|ADT 746x Thermostat");
				i2c->driver = strdup("therm_adt746x");
				i2c->next = devlist;
				devlist = (struct device *)i2c;
				break;
			}
		}
		free_find_results(&list);

		minifind("/proc/device-tree", "mdio", &list);
		// Supported
		for (n = list.result; n; n = n->next) {
			char *matches[] = {"gpio-mdio", NULL};
			if (check_of_compatible(n->path, matches)) {
				i2c = macioNewDevice(NULL);
				i2c->type = CLASS_OTHER;
				i2c->desc = strdup("P.A. Semi Inc.|GPIO MDIO");
				i2c->driver = strdup("gpio_mdio");
				i2c->next = devlist;
				devlist = (struct device *)i2c;
				break;
			}
		}
		free_find_results(&list);
	
	}

	return devlist;
}

#else

struct device *macioProbe( enum deviceClass probeClass, int probeFlags,
		    struct device *devlist ) {
    return devlist;
}

#endif
